structure Interpreter :> Interpreter =
struct

  exception RunError of string*Syntax.pos

  type value = int list

  val seed = ref 3415176
  val seed2 = ref 4151763

  fun rand n =
    let
      val s1 = 73 * !seed + 26111
      val s2 = s1 mod 973177
      val max = (973177 div n) * n
      val s3 = 61 * !seed2 + 26111
      val s4 = s3 mod 990001
      val max2 = (990001 div n) * n
    in
      (seed := s2;
       seed2 := s4;
       if s2>=max orelse s4>=max2 then rand n
       else (s2+s4) mod n + 1)
    end handle Overflow => raise RunError ("rand overflow",(n,!seed))

  fun merge [] l2 = l2
    | merge (a::l1) l2 = merge1 a l1 l2

  and merge1 a l1 [] = a :: l1
    | merge1 a l1 (b::l2) =
        if a<=b then a :: merge1 b l2 l1
	else b :: merge1 a l1 l2

  fun rollDice dice = evalExp dice []

  and evalExp exp table =
    case exp of
      Syntax.NUM (n,p) => [n]
    | Syntax.ID (x,p) =>
        (case lookup x table of
	   SOME v => v
	 | NONE => raise RunError ("unknown variable: "^x,p))
    | Syntax.PLUS (e1,e2,p) =>
        (case (evalExp e1 table, evalExp e2 table) of
	   ([n1], [n2]) => [n1+n2]
	 | _ => raise RunError ("non-singleton arg to +",p))
    | Syntax.MINUS (e1,e2,p) =>
        (case (evalExp e1 table, evalExp e2 table) of
	   ([n1], [n2]) => [n1-n2]
	 | _ => raise RunError ("non-singleton arg to -",p))
    | Syntax.UMINUS (e1,p) =>
        (case (evalExp e1 table) of
	   [n1] => [~n1]
	 | _ => raise RunError ("non-singleton arg to -",p))
    | Syntax.TIMES (e1,e2,p) =>
        (case (evalExp e1 table, evalExp e2 table) of
	   ([n1], [n2]) => [n1*n2]
	 | _ => raise RunError ("non-singleton arg to *",p))
    | Syntax.DIVIDE (e1,e2,p) =>
        (case (evalExp e1 table, evalExp e2 table) of
	   ([], [n2]) => [0]
	 | ([n1], [n2]) =>
	      if n2=0 then raise RunError ("division by 0",p)
	      else [n1 div n2]
	 | _ => raise RunError ("non-singleton arg to /",p))
    | Syntax.CONC (e1,e2,p) =>
        merge (evalExp e1 table) (evalExp e2 table)
    | Syntax.D (e1,p) =>
        (case (evalExp e1 table) of
	   [n]   =>
	      if n<=0 then raise RunError ("Arg to d or D most be >0",p)
	      else [rand n]
	 | _ => raise RunError ("non-singleton arg to d or D",p))
    | Syntax.SUM (e1,p) =>
	[List.foldl (op +) 0 (evalExp e1 table)]
    | Syntax.COUNT (e1,p) =>
	[List.length (evalExp e1 table)]
    | Syntax.LEAST (e1,e2,p) =>
        (case (evalExp e1 table,evalExp e2 table) of
	   ([n], l) =>
	      if n<0 then raise RunError ("Negative arg to least",p)
	      else if List.length l <= n then l
	      else List.take (l, n)
	 | _ => raise RunError ("non-singleton arg to least",p))
    | Syntax.LARGEST (e1,e2,p) =>
        (case (evalExp e1 table,evalExp e2 table) of
	   ([n], l) =>
	      if n<0 then raise RunError ("Negative arg to largest",p)
	      else if List.length l <= n then l
	      else List.drop (l, List.length l - n)
	 | _ => raise RunError ("non-singleton arg to largest",p))
    | Syntax.HASH (e1,e2,p) =>
        (case (evalExp e1 table) of
	   [n] =>
	      if n<0 then raise RunError ("Negative arg to #",p)
	      else
	        foldr (fn (a,b) => merge a b) []
	              (List.tabulate (n,(fn x => evalExp e2 table)))
	 | _ => raise RunError ("non-singleton arg to #",p))
    | Syntax.EQ (e1,e2,p) =>
        (case (evalExp e1 table, evalExp e2 table) of
	   ([n1], l) => List.filter (fn x=>x=n1) l
	 | _ => raise RunError ("non-singleton arg to =",p))
    | Syntax.LT (e1,e2,p) =>
        (case (evalExp e1 table, evalExp e2 table) of
	   ([n1], l) => List.filter (fn x=>x<n1) l
	 | _ => raise RunError ("non-singleton arg to <",p))
    | Syntax.GT (e1,e2,p) =>
        (case (evalExp e1 table, evalExp e2 table) of
	   ([n1], l) => List.filter (fn x=>x>n1) l
	 | _ => raise RunError ("non-singleton arg to >",p))
    | Syntax.LE (e1,e2,p) =>
        (case (evalExp e1 table, evalExp e2 table) of
	   ([n1], l) => List.filter (fn x=>x<=n1) l
	 | _ => raise RunError ("non-singleton arg to <=",p))
    | Syntax.GE (e1,e2,p) =>
        (case (evalExp e1 table, evalExp e2 table) of
	   ([n1], l) => List.filter (fn x=>x>=n1) l
	 | _ => raise RunError ("non-singleton arg to >=",p))
    | Syntax.FROMTO (e1,e2,p) =>
        (case (evalExp e1 table, evalExp e2 table) of
	   ([n1], [n2]) => List.tabulate (n2-n1+1,fn x=>x+n1)
	 | _ => raise RunError ("non-singleton arg to ..",p))
    | Syntax.LET (x,e1,e2,p) =>
        evalExp e2 ((x,evalExp e1 table)::table)
    | Syntax.REPEAT (x,e1,e2,p) =>
        List.foldr (fn (a,b) => merge a b) []
		   (iterate (evalExp e1 table) x e2 table)
    | Syntax.FOREACH (x,e1,e2,p) =>
        foldr (fn (a,b) => merge a b) []
	      (List.map (fn v => evalExp e2 ((x,[v])::table))
			(evalExp e1 table))
    | Syntax.IF (e1,e2,e3,p) =>
        if null (evalExp e1 table)
	then evalExp e3 table
	else evalExp e2 table

  and iterate [] x e table = []
    | iterate v1 x e table =
        v1 :: iterate (evalExp e ((x,v1)::table)) x e table

  and lookup x [] = NONE
    | lookup x ((y,v)::table) =
        if x=y then SOME v else lookup x table

end
